from typing import Any, Callable, Dict, List, Literal, Tuple, Union

from prompt_toolkit.formatted_text import FormattedText

NoneType = type(None)
ARGS_AND_KWARGS = Tuple[Tuple[Any, ...], Dict[str, Any]]
CALLABLE_ERR_HANDLER = Callable[[str, BaseException, str, str], Any]
PromptLikeMsg = Union[str, FormattedText, Callable[..., Union[FormattedText, List[Tuple[str, str]]]]]
LITERAL_EVENTS = Literal[
    "start", "close", "kb_interrupt", "eof", "command_not_found", "command_complete", "command_suggest"
]



from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Union

from prompt_toolkit.completion import Completer, Completion, WordCompleter
from prompt_toolkit.completion.base import CompleteEvent
from prompt_toolkit.completion.nested import NestedDict
from prompt_toolkit.document import Document

from eggella.exceptions import CommandNotFoundError

if TYPE_CHECKING:
    from eggella.manager import CommandManager


class NestedCommandCompleter(Completer):
    """Modification Nested completer with meta information"""

    def __init__(
        self, options: Dict[str, Union[Completer, None]], ignore_case: bool = True, meta: Optional[Dict] = None
    ) -> None:
        self.options = options
        self.meta = meta or {}
        self.ignore_case = ignore_case

    def __repr__(self) -> str:
        return f"NestedCommandCompleter({self.options!r}, {self.meta!r} ignore_case={self.ignore_case!r})"

    @classmethod
    def from_nested_dict(cls, data: NestedDict, meta: Optional[Dict[str, str]] = None) -> "NestedCommandCompleter":
        options: Dict[str, Union[Completer, None]] = {}
        meta = meta or {}
        for key, value in data.items():
            if isinstance(value, Completer):
                options[key] = value
            elif isinstance(value, dict):
                options[key] = cls.from_nested_dict(value, meta)
            elif isinstance(value, set):
                options[key] = cls.from_nested_dict({item: None for item in value}, meta)
            else:
                assert value is None
                options[key] = None
        return cls(options=options, meta=meta)

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        # Split document.
        text = document.text_before_cursor.lstrip()
        stripped_len = len(document.text_before_cursor) - len(text)

        # If there is a space, check for the first term, and use a
        # subcompleter.
        if " " in text:
            first_term = text.split()[0]
            completer = self.options.get(first_term)

            # If we have a sub completer, use this for the completions.
            if completer is not None:
                remaining_text = text[len(first_term) :].lstrip()
                move_cursor = len(text) - len(remaining_text) + stripped_len

                new_document = Document(
                    remaining_text,
                    cursor_position=document.cursor_position - move_cursor,
                )

                yield from completer.get_completions(new_document, complete_event)

        # No space in the input: behave exactly like `WordCompleter`.
        else:
            completer = WordCompleter(list(self.options.keys()), ignore_case=self.ignore_case, meta_dict=self.meta)
            yield from completer.get_completions(document, complete_event)


class CommandCompleter(Completer):
    def __init__(self, manager: "CommandManager", ignore_case: bool = True):
        self.manager = manager
        self.ignore_case = ignore_case

    def get_completions(self, document: Document, complete_event: CompleteEvent):
        text_before_cursor = document.text_before_cursor
        if self.ignore_case:
            text_before_cursor = text_before_cursor.lower()
        text_arr = text_before_cursor.split(" ")
        last_words = text_arr[-1]
        completions = self.__get_current_completions(text_arr[:-1])
        if all(isinstance(d, dict) for d in completions):
            try:
                nested, meta = completions
                yield from NestedCommandCompleter.from_nested_dict(nested, meta).get_completions(
                    document, complete_event
                )
            except (ValueError, TypeError):  # unpack err
                yield
        else:
            # echo({'echo': None}, {})
            for completion, meta in completions:
                if completion not in document.text_before_cursor and "=" not in last_words:
                    yield Completion(completion, -len(last_words), display_meta=meta or "")

    def __get_current_completions(self, text_arr: List[str]):
        if not text_arr:
            return self.manager.all_completions
        command = text_arr[0]
        try:
            command_obj = self.manager.get(command)
        except CommandNotFoundError:
            return []
        if command_obj.nested_completions:
            return command_obj.nested_completions, command_obj.nested_meta
        elif command_obj:
            return [command_obj.completion]



from typing import Any, Callable, Dict, List, Optional, Tuple

from eggella.command.abc import ABCCommandHandler
from eggella.command.arg_caster import CommandArgumentsCaster
from eggella.command.parser import TokensParser, TokensParserRaw

ARGS_AND_KWARGS = Tuple[Tuple[Any], Dict[str, Any]]


class CommandHandler(ABCCommandHandler):
    """default command argument. Split args by `shlex.split` function and cast type from function annotations"""

    def __init__(
        self,
        tokenizer: Callable[[str], List[str]] = TokensParser(),
        caster: Optional[Callable[[Callable, List[str]], ARGS_AND_KWARGS]] = CommandArgumentsCaster(),
    ):
        self.tokenizer = tokenizer
        self.caster = caster

    def handle(self, fn: Callable[..., Any], text: str) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        arguments = self.tokenizer(text)
        return self.caster(fn, arguments) if self.caster else (tuple(arguments), {})

    def __call__(self, fn: Callable[..., Any], text: str):
        return self.handle(fn, text)


class RawCommandHandler(CommandHandler):
    """disable arguments parse and return raw string"""

    def __init__(self, tokenizer=TokensParserRaw(), caster=None):
        super().__init__(tokenizer, caster)



import shlex
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Type,
)

from prompt_toolkit.completion.nested import NestedDict

from eggella._types import CALLABLE_ERR_HANDLER
from eggella.command.completer import CommandCompleter
from eggella.command.handler import CommandHandler
from eggella.command.objects import Command
from eggella.events.events import (
    OnCommandArgumentValueError,
    OnCommandCompleteSuccess,
    OnCommandError,
    OnCommandNotFound,
    OnCommandTooManyArgumentsError,
    OnEOFError,
    OnFSMEOFError,
    OnFSMKeyboardInterrupt,
    OnKeyboardInterrupt,
    OnSuggest,
)
from eggella.exceptions import (
    CommandArgumentValueError,
    CommandNotFoundError,
    CommandTooManyArgumentsError,
)
from eggella.shortcuts.help_pager import gen_help_commands, gen_man_pager

if TYPE_CHECKING:
    from eggella.app import Eggella


_ErrorEventsMapping = Dict[str, Tuple[Type[BaseException], ...]]


class CommandManager:
    def __init__(self, app: "Eggella"):
        self._app = app
        self.commands: Dict[str, Command] = {}
        self.error_events: Dict[str, Callable[[str, BaseException, str, str], Any]] = {}
        self.handled_exceptions: _ErrorEventsMapping = {}

    @staticmethod
    def _simple_parse_arguments(raw_command: str) -> Tuple[Tuple[str, ...], Dict[str, str]]:
        tokens = shlex.split(raw_command)
        args: List[str] = []
        kwargs: Dict[str, str] = {}
        for token in tokens:
            if "=" in token:
                key, value = token.split("=", 1)
                kwargs[key] = value
            else:
                args.append(token)
        return tuple(args), kwargs

    def exec(self, key: str, args: str):
        command = self.get(key)

        if not command.is_visible:
            raise CommandNotFoundError

        try:
            return command.handle(args)
        except TypeError as e:
            if 'too many positional arguments' in e.args[0]:
                raise CommandTooManyArgumentsError("Too many arguments") from e
            else:
                raise e
        except ValueError as e:
            raise CommandArgumentValueError(f"Wrong argument type passed: {e.args}") from e
        except BaseException as e:
            if err_handler := self.error_events.get(command.fn.__name__):
                handle_exceptions = self.handled_exceptions.get(command.fn.__name__, None)
                if handle_exceptions and any(e.__class__ == exc for exc in handle_exceptions):
                    _args, _kwargs = self._simple_parse_arguments(args)
                    return err_handler(key, e, *_args, **_kwargs)
                else:
                    raise e
            raise e

    def get_completer(self) -> CommandCompleter:
        return CommandCompleter(self)  # type: ignore

    def on_error(self, *errors: Type[BaseException]):
        def decorator(handler: CALLABLE_ERR_HANDLER):
            @wraps(handler)
            def decorator_wrapper(func: Callable[..., Any]):
                if not self.error_events.get(func.__name__):
                    self.error_events[func.__name__] = handler
                    self.handled_exceptions[func.__name__] = errors

                @wraps(func)
                def wrapper(*args, **kwargs):
                    try:
                        return func(*args, **kwargs)
                    except BaseException as e:
                        if any(e.__class__ == exc for exc in errors):
                            return handler("", e, *args, **kwargs)
                        raise e

                return wrapper

            return decorator_wrapper

        return decorator

    @property
    def all_completions(self) -> List[Tuple[str, str]]:
        return [com.completion for com in self.commands.values() if com.is_visible]

    def get(self, key: str) -> Command:
        if command := self.commands.get(key, None):
            return command
        raise CommandNotFoundError(f"Command {key} not founded")

    def command(
        self,
        key: Optional[str] = None,
        *,
        short_description: Optional[str] = None,
        usage: Optional[str] = None,
        cmd_handler: Optional[Callable[[Callable[..., Any], str], Tuple[Tuple[Any, ...], Dict[str, Any]]]] = None,
        nested_completions: Optional[NestedDict] = None,
        nested_meta: Optional[Dict[str, Any]] = None,
        is_visible: bool = True,
    ):
        def decorator(func: Callable):
            self.register_command(
                func,
                key=key,
                short_description=short_description,
                usage=usage,
                cmd_handler=cmd_handler,
                nested_completions=nested_completions,
                nested_meta=nested_meta,
                is_visible=is_visible,
            )

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def register_command(
        self,
        func: Callable,
        key: Optional[str] = None,
        *,
        short_description: Optional[str] = None,
        usage: Optional[str] = None,
        cmd_handler: Optional[Callable[[Callable[..., Any], str], Tuple[Tuple[Any, ...], Dict[str, Any]]]] = None,
        nested_completions: Optional[NestedDict] = None,
        nested_meta: Optional[Dict[str, Any]] = None,
        is_visible: bool = True,
    ):
        if not key:
            key = func.__name__
        if self.commands.get(key):
            raise TypeError(f"Command '{key}' already register")

        if not cmd_handler:
            self.commands[key] = Command(
                fn=func,
                key=key,
                handler=CommandHandler(),
                short_description=short_description,
                usage=usage,
                nested_completions={key: nested_completions},
                nested_meta=nested_meta or {},
                is_visible=is_visible,
            )
        else:
            self.commands[key] = Command(
                fn=func,
                key=key,
                handler=cmd_handler,
                short_description=short_description,
                usage=usage,
                nested_completions={key: nested_completions},
                nested_meta=nested_meta or {},
                is_visible=is_visible,
            )

    def _help_command(self, key: Optional[str] = None):
        """show help or print all available commands if not argument passed"""
        if not key:
            return gen_help_commands(self._app)
        elif comma := self.commands.get(key):
            return f"{key} {comma.command_description}"
        else:
            raise CommandNotFoundError

    def _man_page(self):
        """generate man page view with all commands"""
        gen_man_pager(self._app)

    @staticmethod
    def _exit_command():
        """exit from this application"""
        raise KeyboardInterrupt

    def register_buildin_commands(self):
        self.register_command(self._exit_command, "exit")
        self.register_command(self._man_page, ".man")

        # generate nested for help command
        _nested_commands = {k: None for k in self.commands.keys()}
        _nested_meta = {}
        _nested_meta.update({k: v.command_description for k, v in self.commands.items()})
        self.register_command(
            self._help_command,
            "help",
            usage="help; help exit",
            nested_completions=_nested_commands,
            nested_meta=_nested_meta,
        )


class EventManager:
    def __init__(self, app: "Eggella"):
        self.app = app
        self.startup_events: List[Callable] = []
        self.close_events: List[Callable] = []
        self.errors_events: Dict[str, Callable] = {}
        # TODO typing more accurately
        # loop events
        self.kb_interrupt_event: Callable[..., bool] = OnKeyboardInterrupt()
        self.eof_event: Callable[..., bool] = OnEOFError()
        # commands events
        self.command_error_event: Callable[..., None] = OnCommandError()
        self.command_not_found_event: Callable[..., None] = OnCommandNotFound()
        self.command_complete_event: Callable[..., None] = OnCommandCompleteSuccess()
        self.command_suggest_event: Optional[Callable[..., None]] = OnSuggest()
        self.command_many_args_err_event: Callable[..., None] = OnCommandTooManyArgumentsError()
        self.command_argument_value_err_event: Callable[..., None] = OnCommandArgumentValueError()
        # FSM events
        self.fsm_kb_interrupt_event: Callable[..., bool] = OnFSMKeyboardInterrupt()
        self.fsm_eof_error_event: Callable[..., bool] = OnFSMEOFError()

    def register_event(
        self,
        name: Literal[
            "start", "close", "kb_interrupt", "eof", "command_not_found", "command_complete", "command_suggest"
        ],
        func: Optional[Callable],
    ):
        if name == "start" and func:
            self.startup_events.append(func)
            return
        elif name == "close" and func:
            self.close_events.append(func)
            return
        elif name == "kb_interrupt" and func:
            self.kb_interrupt_event = func
            return
        elif name == "eof" and func:
            self.eof_event = func
            return
        elif name == "command_not_found" and func:
            self.command_not_found_event = func
            return
        elif name == "command_complete" and func:
            self.command_complete_event = func
            return
        elif name == "command_suggest":
            self.command_suggest_event = func
            return
        raise TypeError("Unknown event type")

    def startup(self):
        def decorator(func: Callable):
            self.register_event("start", func)

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def close(self):
        def decorator(func: Callable):
            self.register_event("close", func)

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator


class BlueprintManager:
    def __init__(self, main_app: "Eggella"):
        self.app = main_app

        self.blueprints: List["Eggella"] = []
        self._loaded_blueprints: Set[str] = set()

    def register_blueprints(self, *bp_apps: "Eggella"):
        for bp_app in bp_apps:
            self.blueprints.append(bp_app)

    def load_blueprints(self):
        for blueprint in self.blueprints:
            if blueprint.app_name in self._loaded_blueprints:
                continue
            # register commands to main app
            for key, command in blueprint.command_manager.commands.items():
                if self.app.command_manager.commands.get(key) and not self.app.overwrite_commands_from_blueprints:
                    raise TypeError(
                        f"Command '{key}' from blueprint `{blueprint.app_name}` already registered. "
                        f"If you need overwrite commands set `overwrite_commands_from_blueprints=True`"
                    )
                self.app.command_manager.commands[key] = command
            # register FSM groups to main app
            for key, fsm_state in blueprint.fsm.fsm_storage.items():
                self.app.fsm.fsm_storage[key] = fsm_state

            # register events
            for start_ev in blueprint.event_manager.startup_events:
                self.app.event_manager.startup_events.append(start_ev)

            for close_ev in blueprint.event_manager.close_events:
                self.app.event_manager.close_events.append(close_ev)

            self._loaded_blueprints.add(blueprint.app_name)



class BaseEgellaException(Exception):
    pass


class CommandNotFoundError(BaseEgellaException):
    pass


class CommandParseError(BaseEgellaException):
    pass


class CommandTooManyArgumentsError(BaseEgellaException):
    pass


class CommandArgumentValueError(BaseEgellaException):
    pass



import re
from typing import Iterable

from prompt_toolkit.completion import CompleteEvent, Completion
from prompt_toolkit.completion import FuzzyCompleter as OldFuzzyCompleter
from prompt_toolkit.completion.fuzzy_completer import _FuzzyMatch
from prompt_toolkit.document import Document


class FuzzyCompleter(OldFuzzyCompleter):
    """Patched original FuzzyCompleter for avoid
    AttributeError: 'NoneType' object has no attribute 'text'
    in prompt for next cases:

    1. tap whitespace char only

    2. tap whitespace char + any text
    """

    def _get_fuzzy_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        word_before_cursor = document.get_word_before_cursor(pattern=re.compile(self._get_pattern()))

        # Get completions
        document2 = Document(
            text=document.text[: document.cursor_position - len(word_before_cursor)],
            cursor_position=document.cursor_position - len(word_before_cursor),
        )

        inner_completions = list(self.completer.get_completions(document2, complete_event))

        fuzzy_matches: list[_FuzzyMatch] = []  # type: ignore

        if word_before_cursor == "":
            # If word before the cursor is an empty string, consider all
            # completions, without filtering everything with an empty regex
            # pattern.
            fuzzy_matches = [_FuzzyMatch(0, 0, compl) for compl in inner_completions]
        else:
            pat = ".*?".join(map(re.escape, word_before_cursor))
            pat = f"(?=({pat}))"  # lookahead regex to manage overlapping matches
            regex = re.compile(pat, re.IGNORECASE)
            for compl in inner_completions:
                # first path
                if compl is None:
                    continue
                matches = list(regex.finditer(compl.text))
                if matches:
                    # Prefer the match, closest to the left, then shortest.
                    best = min(matches, key=lambda m: (m.start(), len(m.group(1))))
                    fuzzy_matches.append(_FuzzyMatch(len(best.group(1)), best.start(), compl))

            def sort_key(fuzzy_match: _FuzzyMatch) -> tuple[int, int]:  # type: ignore
                "Sort by start position, then by the length of the match."
                return fuzzy_match.start_pos, fuzzy_match.match_length

            fuzzy_matches = sorted(fuzzy_matches, key=sort_key)

        for match in fuzzy_matches:
            if match.completion is None:
                # second path
                yield Completion("")
            else:
                # Include these completions, but set the correct `display`
                # attribute and `start_position`.
                yield Completion(
                    text=match.completion.text,
                    start_position=match.completion.start_position - len(word_before_cursor),
                    # We access to private `_display_meta` attribute, because that one is lazy.
                    display_meta=match.completion._display_meta,
                    display=self._get_display(match, word_before_cursor),
                    style=match.completion.style,
                )



from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Literal,
    Optional,
    Type,
    Union,
    overload,
)

from prompt_toolkit import HTML, PromptSession
from prompt_toolkit.completion.nested import NestedDict

from eggella._patches import FuzzyCompleter
from eggella._types import ARGS_AND_KWARGS, LITERAL_EVENTS, PromptLikeMsg
from eggella.command.abc import ABCCommandHandler
from eggella.command.objects import Command
from eggella.exceptions import (
    CommandArgumentValueError,
    CommandNotFoundError,
    CommandParseError,
    CommandTooManyArgumentsError,
)
from eggella.fsm.fsm import FsmController, IntStateGroup
from eggella.manager import BlueprintManager, CommandManager, EventManager
from eggella.shortcuts.cmd_shortcuts import CmdShortCuts

_DEFAULT_INTRO_MSG = HTML(
    "<ansired>Press |CTRL+C| or |CTRL+D| or type</ansired> exit <ansired>for close this app</ansired>\n"
    "<ansigreen>Press |TAB| or type</ansigreen> help <ansigreen>for get commands information</ansigreen>"
)


class Eggella:
    __app_instances__: Dict[str, "Eggella"] = {}

    def __new__(cls, app_name: str, msg: PromptLikeMsg = "> ") -> "Eggella":
        if _app := cls.__app_instances__.get(app_name):
            return _app

        new_app = super().__new__(cls)
        cls.__app_instances__[app_name] = new_app
        return new_app

    def __init__(self, app_name: str, msg: PromptLikeMsg = "> "):
        self.app_name = app_name
        self.prompt_msg = msg
        self.session: PromptSession = PromptSession(msg)
        self.cmd = CmdShortCuts()

        # app context storage
        self.CTX: Dict[Hashable, Any] = {}
        # config
        self._intro: Union[HTML, PromptLikeMsg] = _DEFAULT_INTRO_MSG
        self._doc: str = ""
        self.overwrite_commands_from_blueprints: bool = False

        # managers
        self._command_manager: CommandManager = CommandManager(self)
        self._event_manager = EventManager(self)
        self._blueprint_manager = BlueprintManager(self)

        # fsm
        self.fsm = FsmController(self)

    @property
    def blueprint_manager(self):
        """Get blueprint manager"""
        return self._blueprint_manager

    @property
    def command_manager(self) -> CommandManager:
        """Get command manager"""
        return self._command_manager

    @property
    def event_manager(self) -> EventManager:
        """Get event manager"""
        return self._event_manager

    @property
    def documentation(self):
        """Get full manual text for help render"""
        return self._doc

    @documentation.setter
    def documentation(self, text: str):
        self._doc = text

    @property
    def intro(self):
        """startup intro text"""
        return self._intro

    @intro.setter
    def intro(self, text: Union[HTML, PromptLikeMsg]):
        self._intro = text

    def on_startup(self):
        """Register event manager on startup app"""
        return self.event_manager.startup()

    def on_close(self):
        """Register event manager on close app"""
        return self.event_manager.close()

    def on_error(self, *errors: Type[BaseException]):
        """Create error handler

        :param errors: types of Exceptions to be handled
        """
        return self.command_manager.on_error(*errors)

    def on_command(
        self,
        key: Optional[str] = None,
        short_description: Optional[str] = None,
        *,
        usage: Optional[str] = None,
        cmd_handler: Optional[Callable[[Callable[..., Any], str], ARGS_AND_KWARGS]] = None,
        nested_completions: Optional[NestedDict] = None,
        nested_meta: Optional[Dict[str, Any]] = None,
        is_visible: bool = True,
    ):
        """Register command

        :param key: command key. if not set - set key as function name
        :param short_description: short description
        :param usage: usage example
        :param cmd_handler: Command handler
        :param nested_completions: nested completer
        :param nested_meta: nested meta information
        :param is_visible: set visible command
        """
        return self._command_manager.command(
            key,
            short_description=short_description,
            usage=usage,
            cmd_handler=cmd_handler,
            nested_completions=nested_completions,
            nested_meta=nested_meta,
            is_visible=is_visible,
        )

    def on_state(self, state: IntStateGroup):
        """register state handler

        :param state: IntStateGroup state
        """
        return self.fsm.state(state)

    def register_states(self, states: Type[IntStateGroup]):
        """Register group states in this application

        :param states: IntStateGroup class
        :return:
        """
        self.fsm.attach(states)

    def register_blueprint(self, *apps: "Eggella"):
        """register blueprint for extension. Add on_startup, on_close, on_command, on_state events

        :param apps: Eggella applications
        :return:
        """
        self.blueprint_manager.register_blueprints(*apps)

    def _load_blueprints(self):
        self.blueprint_manager.load_blueprints()

    def register_command(
        self,
        func: Callable,
        key: Optional[str] = None,
        short_description: Optional[str] = None,
        *,
        is_visible: bool = True,
        usage: Optional[str] = None,
        cmd_handler: Optional[ABCCommandHandler] = None,
    ):
        """Register command from function"""
        self._command_manager.register_command(
            func, key, short_description=short_description, usage=usage, cmd_handler=cmd_handler, is_visible=is_visible
        )

    @overload
    def register_event(
        self,
        name: Literal["start", "close", "kb_interrupt", "eof", "command_not_found", "command_complete"],
        func: Callable,
    ) -> None:
        pass

    @overload
    def register_event(
        self,
        name: Literal["command_suggest"],
        func: Optional[Callable],
    ) -> None:
        pass

    def register_event(
        self,
        name: LITERAL_EVENTS,
        func: Optional[Callable],
    ) -> None:
        self._event_manager.register_event(name, func)

    def has_command(self, key: str) -> bool:
        """return True if command is registered

        :param key: command key
        :return:
        """
        return bool(self.command_manager.commands.get(key, None))

    def remove_command(self, key: str):
        """remove command from this application. raise KeyError, if command not founded

        :param key: command key
        :return:
        """
        if self.has_command(key):
            self.command_manager.commands.pop(key)
        else:
            raise KeyError

    def loop(self):
        """Run this application"""
        self.cmd.print_ft(self.intro)
        self._load_blueprints()
        self._command_manager.register_buildin_commands()
        self._handle_startup_events()
        self._handle_commands()
        self._handle_close_events()

    def _handle_startup_events(self):
        for event in self._event_manager.startup_events:
            event()

    def _handle_close_events(self):
        for event in self._event_manager.close_events:
            event()

    def get_command(self, key: str) -> Command:
        """get command object from command_manager"""
        return self.command_manager.get(key)

    def _handle_error(self, key, args, exc):
        try:
            raise exc
        except CommandNotFoundError:
            self.event_manager.command_not_found_event(key, args)
            if self.event_manager.command_suggest_event:
                self.event_manager.command_suggest_event(
                    key, [c.key for c in self._command_manager.commands.values() if c.is_visible]
                )
        except CommandParseError:
            self.event_manager.command_error_event(key, args)
        except CommandTooManyArgumentsError as exc:
            self.event_manager.command_many_args_err_event(key, args, exc)
        except CommandArgumentValueError as exc:
            self.event_manager.command_argument_value_err_event(key, args, exc)

    def _handle_commands(self):
        """application loop"""
        while True:
            try:
                # if FSM activated - handle this
                if self.fsm.is_active():
                    # handle fsm events
                    try:
                        self.fsm.current()
                        continue
                    except KeyboardInterrupt:
                        if self.event_manager.fsm_kb_interrupt_event():
                            self.fsm.finish()
                    except EOFError:
                        if self.event_manager.fsm_eof_error_event():
                            self.fsm.finish()
                # handle main app input
                completer = FuzzyCompleter(completer=self.command_manager.get_completer())
                result = self.session.prompt(self.prompt_msg, completer=completer)
                if not result:
                    continue

                if (tokens := result.split(" ", 1)) and len(tokens) == 1:
                    key, args = tokens[0], ""
                else:
                    key, args = tokens[0], tokens[1]
                # handle input command
                if result := self.command_manager.exec(key, args):
                    self.event_manager.command_complete_event(result)
            # exit exceptions
            except KeyboardInterrupt:
                if self.event_manager.kb_interrupt_event():
                    break
            except EOFError:
                if self.event_manager.eof_event():
                    break
            # command errors handle
            except Exception as exc:
                self._handle_error(key, args, exc)



import inspect
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Tuple

from prompt_toolkit.completion.nested import NestedDict

from eggella._types import ARGS_AND_KWARGS
from eggella.command.handler import CommandHandler


@dataclass
class Command:
    fn: Callable[..., Any]
    key: str
    handler: Callable[[Callable[..., Any], str], ARGS_AND_KWARGS] = CommandHandler()
    usage: Optional[str] = None
    short_description: Optional[str] = None
    nested_completions: Optional[NestedDict] = None
    nested_meta: Dict[str, Any] = field(default_factory=dict)
    is_visible: bool = True

    def handle(self, command_text: str) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        args, kwargs = self.handler(self.fn, command_text)
        return self.fn(*args, **kwargs)

    @property
    def arguments(self) -> List[str]:
        args = inspect.signature(self.fn).parameters.values()
        return [str(arg) for arg in args]

    @property
    def docstring(self) -> str:
        return inspect.getdoc(self.fn) or ""

    def get_short_description(self):
        if self.short_description:
            return self.short_description
        elif self.docstring:
            return self.docstring.split("\n")[0].strip()
        return ""

    @property
    def command_description(self):
        arg_list = ", ".join(f"{arg}" for arg in self.arguments)
        if self.short_description:
            if arg_list:
                return f"({arg_list}) - {self.short_description}"
            return self.short_description
        elif self.docstring:
            short_desc = self.docstring.split("\n")[0]
            return f"({arg_list}) - {short_desc}" if arg_list else short_desc
        return f"({arg_list})"

    @property
    def completion(self) -> Tuple[str, str]:
        return self.key, self.command_description

    @property
    def help(self):
        arg_list = " ".join(f"[{arg}]" for arg in self.arguments)
        if len(self.docstring.split("\n")) > 1:
            if self.usage:
                return f"      {self.key} ({arg_list})\n            {self.docstring}\nUSAGE:\n      {self.usage}"
            return f"      {self.key} ({arg_list})\n            {self.docstring}"
        if self.usage:
            return f"      {self.key} ({arg_list}) - {self.docstring}\nUSAGE:\n      {self.usage}"
        return f"      {self.key} ({arg_list}) - {self.docstring}"



import shlex
import warnings
from typing import List

from eggella.command.abc import ABCTokensParser


class TokensParser(ABCTokensParser):
    """Default tokens parser. Split input string arguments by `shlex.split` function"""

    def __call__(self, raw_command: str) -> List[str]:
        return shlex.split(raw_command)


class TokensParserRaw(ABCTokensParser):
    """Tokens split input string to arguments"""

    def __call__(self, raw_command: str) -> List[str]:
        return [raw_command]


class CommandParser(TokensParser):
    def __init__(self):
        warnings.warn(
            "CommandParser has deprecated, please use `TokenParser` instead", stacklevel=3, category=DeprecationWarning
        )


class CommandParserRaw(TokensParserRaw):
    def __init__(self):
        warnings.warn(
            "CommandParserRaw has deprecated, please use `TokenParserRaw` instead",
            stacklevel=3,
            category=DeprecationWarning,
        )
